/**********************************************************************/
/*                   I R Q U T I L  . C                               */
/**------------------------------------------------------------------**/
/*  Aufgabe        : Stellt Funktionen fr Hardware-Interrupt         */
/*                   Programmierung zur Verfgung.                    */
/**------------------------------------------------------------------**/
/*  Autor          : Michael Tischer / Bruno Jennrich                 */
/*  entwickelt am  : 12.03.1994                                       */
/*  letztes Update : 26.03.1994                                       */
/**********************************************************************/
#ifndef __IRQUTIL_C                     /* Kann auch #Included werden */
#define __IRQUTIL_C

/*-- Include-Dateien einbinden ---------------------------------------*/
#include <dos.h>
#include <conio.h>

#include "irqutil.h"

/**********************************************************************/
/* irq_Enable : Hardware-Interrupt zulassen                           */
/**------------------------------------------------------------------**/
/* Eingabe : iIRQ - Nummer des Hardware-Interrupts (0-15) der zu-     */
/*                  gelassen werden soll                              */
/**------------------------------------------------------------------**/
/* Info : Das An- und Abschalten von IRQ's geschieht durch            */
/*        bermittlung eines OCW1 (Operation Control Word) genannten  */
/*        Bytes an den Basisport des IRQ-Controllers.                 */
/**********************************************************************/
VOID irq_Enable( INT iIRQ )
{ INT iPort;          /* Port Adresse des zustndigen PIC's ermitteln */
                           /* ( 0-7 = MASTER_PIC , 8-15 = SLAVE_PIC ) */
  iPort = ( iIRQ <= 7 ) ? MASTER_PIC : SLAVE_PIC ;
  iPort += IRQ_MASK;                    /* Maskierungs-Port auswhlen */

  iIRQ &= 0x0007;              /* PIC-Interruptnummer ermitteln (0-7) */
                               /* Bit Lschen -> Interrupt zugelassen */
  outp( iPort, inp( iPort ) & ~( 1 << iIRQ ) );
}

/**********************************************************************/
/* irq_Disable : Hardware-Interrupt verbieten                         */
/**------------------------------------------------------------------**/
/* Eingabe : iIRQ - Nummer des Hardware-Interrupts (0-15) der ver-    */
/*                  boten werden soll                                 */
/**********************************************************************/
VOID irq_Disable( INT iIRQ )
{ INT iPort;          /* Port Adresse des zustndigen PIC's ermitteln */
                           /* ( 0-7 = MASTER_PIC , 8-15 = SLAVE_PIC ) */
  iPort = ( iIRQ <= 7 ) ? MASTER_PIC : SLAVE_PIC ;
  iPort += IRQ_MASK;                    /* Maskierungs-Port auswhlen */

  iIRQ &= 0x0007;              /* PIC-Interruptnummer ermitteln (0-7) */
                                  /* Bit Setzen -> Interrupt gesperrt */
  outp( iPort, inp( iPort ) | ( 1 << iIRQ ) );
}

/**********************************************************************/
/* irq_SendEOI : "End Of Interrupt" signalisieren                     */
/**------------------------------------------------------------------**/
/* Eingabe : iIRQ : Nummer des Hardware-Interrupts (0-15) der komplet */
/*                  abgearbeitet wurde.                               */
/* Info : - Ein Interrupt kann erst dann wieder ausgelst werden,     */
/*          wenn die vorherige Auslsung vollstndig bearbeitet       */
/*          wurde. Um den PIC's zu signalisieren, das ein Interrupt-  */
/*          handler alle Aufgaben erledigt hat, mu an die PIC's ein  */
/*          EOI gesendet werden.                                      */
/*          EOI - ohne IRQ-Nummer weil im Nested-Mode                 */
/**********************************************************************/
VOID irq_SendEOI( INT iIRQ )
{                        /* Bei IRQ 8 - 15 auch Slave benachrichtigen */
    if ( iIRQ > 7 ) outp( SLAVE_PIC, EOI );
    outp( MASTER_PIC, EOI );    /* Dem MASTER immer EOI signalisieren */
}

/**********************************************************************/
/* irq_SetHandler : Neuen Interrupt-Handler installieren              */
/**------------------------------------------------------------------**/
/* Eingabe : iIRQ      - Nummer des Interrupts, der neuen Handler     */
/*                       erhalten soll.                               */
/*           lpHandler - Adresse des neuen Handlers                   */
/* Ausgabe : Adresse des alten Handlers.                              */
/**********************************************************************/
VOID ( _interrupt _FP *irq_SetHandler( INT iIRQ,
                     VOID (_interrupt _FP *lpHandler)() ) ) ()
{ VOID (_interrupt _FP *lpOldHandler)();
  INT  iVect;
                 /* Interupt-Vektor des Hardware-Interrupts ermitteln */
                 /* IRQ 0 - 7  = Vektoren 0x08 - 0x0F                 */
                 /* IRQ 8 - 15 = Vektoren 0x70 - 0x77                 */
  iVect = ( iIRQ <= 7 ) ?
          ( MASTER_FIRST_VECTOR + iIRQ ) :
          ( SLAVE_FIRST_VECTOR + ( iIRQ & 0x0007 ) );

  irq_Disable( iIRQ );   /* Hardware- und Software-Interrupts sperren */
  _disable();

  lpOldHandler = _dos_getvect( iVect );       /* Alten Handler retten */
  _dos_setvect( iVect, lpHandler );           /* Neuen Handler setzen */

  _enable();                          /* Software-Interrupts erlauben */
  if( lpHandler )    /* Falls ein Handler bergeben wurde auch wieder */
    irq_Enable( iIRQ ); /* entsprechenden Hardware-Interrupt erlauben */

  return lpOldHandler;      /* Adresse des alten Handlers zurckgeben */
}

/**********************************************************************/
/* irq_ReadMask : Lese Maskierungen eines IRQ-Controllers             */
/**------------------------------------------------------------------**/
/* Eingabe : iController - Basisport des IRQ-Controllers (0x20/0xA0)  */
/* Ausgabe : Maskierung der vom IRQ-Controller bedienten IRQ's.       */
/*           Gesetzte Bits : IRQ wird nicht ausgelst.                */
/**------------------------------------------------------------------**/
/* Info : Die Maskierungen eines PIC's knnen durch einfaches aus-    */
/*        lesen des Mask-Registers ermittelt werden. (0x21/0xA1)      */
/**********************************************************************/
BYTE irq_ReadMask( INT iController )
{
  return ( BYTE )inp( iController + IRQ_MASK );
}

/**********************************************************************/
/* irq_ReadISR : Lese Statusregister eines Irq-Controllers            */
/**------------------------------------------------------------------**/
/* Eingabe : iController - Basisport des IRQ-Controllers (0x20/0xA0)  */
/* Ausgabe : Statusregister der vom IRQ-Controller bedienten IRQ's.   */
/*           Gesetzte Bits : die enstprechende IRQ-Routine wird       */
/*           augenblich ausgefhrt oder durch einen hher prio-       */
/*           risierten IRQ unterbrochen.                              */
/**------------------------------------------------------------------**/
/* Info : Bevor der IRQ-Controller den Inhalt des ISR-Registers       */
/*        mu das OCW3 genannte Byte an den Controller bermittelt    */
/*        werden.                                                     */
/**********************************************************************/
BYTE irq_ReadISR( INT iController )
{              /* OCW3: 0x0B = Keine Operation , kein Poll, ISR lesen */
  outp( iController, 0x0B );
  return ( BYTE )inp( iController );
}

/**********************************************************************/
/* irq_ReadIRR : Lese Requestregister eines Irq-Controllers           */
/**------------------------------------------------------------------**/
/* Eingabe : iController - Basisport des IRQ-Controllers (0x20/0xA0)  */
/* Ausgabe : Requestregister der vom IRQ-Controller bedienten IRQ's.  */
/*           Gesetzte Bits : die enstprechende IRQ-Routine soll als   */
/*           nchstes ausgefhrt werden, sofer kein hher prio-       */
/*           risierter Request ansteht.                               */
/**------------------------------------------------------------------**/
/* Info : Bevor der IRQ-Controller den Inhalt des ISR-Registers       */
/*        mu das OCW3 genannte Byte an den Controller bermittelt    */
/*        werden.                                                     */
/**********************************************************************/
BYTE irq_ReadIRR( INT iController )
{              /* OCW3: 0x0B = Keine Operation , kein Poll, IRR lesen */
  outp( iController, 0x0A );
  return ( BYTE )inp( iController );
}

#endif

