/**********************************************************************/
/*                     S E R T R A N S . C                            */
/**------------------------------------------------------------------**/
/*  Aufgabe        : Filetransfer mit Fehlerkorrektur ber serielle   */
/*                   Schnittstelle                                    */
/**------------------------------------------------------------------**/
/*  Autor          : Michael Tischer / Bruno Jennrich                 */
/*  entwickelt am  : 08.04.1994                                       */
/*  letztes Update : 18.04.1994                                       */
/* COMPILER: Borland C++ 3.1, Microsoft Visual C++ 1.5                */
/**********************************************************************/
#include <dos.h>
#include <stdio.h>
#include <string.h>
#include <process.h>

#include "serutil.h"
#include "sertrans.h"

/*- Folgende Zeilen innerhalb eines Projektes ausdokumentieren -------*/
#include "serutil.c"
#include "win.c"
#include "irqutil.c"
#include "args.c"

/**********************************************************************/
/* CalcCRC : Cyclic-Redundancy-Checksum berechnen                     */
/**------------------------------------------------------------------**/
/* Eingabe : uCRC  - bisherige Checksum                               */
/*           bData - einzubeziehendes Byte                            */
/* Ausgabe : neue Checksumme inkl. des bergebenen Bytes              */
/**------------------------------------------------------------------**/
/* Info : Beim ersten Aufruf dieser Funktion mu die "bisherige"      */
/*        Checksumme auf 0xFFFF gesetzt werden.                       */
/**********************************************************************/
UINT CalcCRC( UINT uCRC, BYTE bData )
{ static INT j;                /* static's sind ein bischen schneller */
  for( j = 0; j < 8; j++ )
  {
    uCRC <<= 1;                             /* Modulo Polynomdivision */
    if( bData & 0x0080 ) uCRC ^= CCITT_POLYNOM;
    bData <<= 1;
  }
  return uCRC;
}

/**********************************************************************/
/* paket_Build : Struktur vom Typ PAKET fr das Senden vorbereiten    */
/**------------------------------------------------------------------**/
/* Eingabe : pPaket    - Adresse des zu initialisierenden Pakets      */
/*           bType     - Benutzertyp des Pakets                       */
/*           pData     - Adresse der Paket-Daten                      */
/*           uDataSize - Anzahl der im Paket enthaltenen gltigen     */
/*                       Bytes                                        */
/*           uNr       - laufende Nummer des Pakets                   */
/**********************************************************************/
VOID paket_Build( PPAKET  pPaket,
                  BYTE    bType,
                  PBYTE   pData,
                  UINT    uDataSize,
                  UINT    uNr )
{ UINT i;
  pPaket->bType = bType;
  pPaket->uDataSize = uDataSize;
  pPaket->uNr = uNr;
  for( i = 0; i != uDataSize; i++ )
       pPaket->bData[ i ] = pData[ i ];
}

/**********************************************************************/
/* trans_WriteUINT : Sende einzelnes UINT ber serielle Schnittstelle */
/**------------------------------------------------------------------**/
/* Eingabe : iSerPort - Basisport der Schnittstelle                   */
/*           uData    - zu versendendes UINT                          */
/* Ausgabe : Fehlerstatus                                             */
/**------------------------------------------------------------------**/
/* Info : Diese Funktion verschickt zunchst das LO-, dann das HI-    */
/*        BYTE eines UINTs. Achtung! Kein Handshake!                  */
/**********************************************************************/
INT trans_WriteUINT( INT iSerPort, UINT uData )
{
  if( ser_WriteByte( iSerPort,
                     LOBYTE( uData ),
                     0x8000,
                     0,
                     0 ) == SER_SUCCESS )
     if( ser_WriteByte( iSerPort,
                        HIBYTE( uData ),
                        0x8000,
                        0,
                        0 ) == SER_SUCCESS )
         return TRANS_SUCCESS;
  return TRANS_TRANSERROR;
}

/**********************************************************************/
/* trans_ReadUINT : Ein einzelnes UINT von Schnittstelle lesen       */
/**------------------------------------------------------------------**/
/* Eingabe : iSerPort - Basisport der Schnittstelle                   */
/*           pData    - Adresse eines UINT's, das empfangenen Wert    */
/*                      aufnehmen soll.                               */
/* Ausgabe : Fehlerstatus                                             */
/**********************************************************************/
INT trans_ReadUINT( INT iSerPort, PUINT pData )
{ BYTE bErrL, bErrH;
  if( ser_ReadByte( iSerPort, &bErrL, 0x8000, 0, 0 ) == SER_SUCCESS )
     if( ser_ReadByte( iSerPort, &bErrH, 0x8000, 0, 0 ) == SER_SUCCESS )
     {
       *pData = MAKEWORD( bErrH, bErrL );
       return TRANS_SUCCESS;
     }
  return TRANS_TRANSERROR;
}

/**********************************************************************/
/* trans_WriteNak : Negative Aknowledge ber Schnittstelle senden     */
/**------------------------------------------------------------------**/
/* Eingabe : iSerPort - Basisport der Schnittstelle                   */
/*           iCode    - erweiterter Fehlercode                        */
/* Ausgabe : Fehlerstatus                                             */
/**********************************************************************/
INT trans_SendNak( INT iSerPort, UINT iCode )
{
  if( ser_WriteByte( iSerPort, ASCII_NAK, 0x8000,0,0) == SER_SUCCESS )
      return trans_WriteUINT( iSerPort, iCode );
  return TRANS_TRANSERROR;
}

/**********************************************************************/
/* trans_WriteAck : (Positives) Aknowledge ber Schnittstelle senden  */
/**------------------------------------------------------------------**/
/* Eingabe : iSerPort - Basisport der Schnittstelle                   */
/* Ausgabe : Fehlerstatus                                             */
/**********************************************************************/
INT trans_SendAck( INT iSerPort )
{
  if( ser_WriteByte( iSerPort, ASCII_ACK, 0x8000,0,0) == SER_SUCCESS )
    return TRANS_SUCCESS;
  return TRANS_TRANSERROR;
}

/**********************************************************************/
/* trans_SendPaket : Ganzes Datenpaket ber Schnittstelle senden      */
/**------------------------------------------------------------------**/
/* Eingabe : iSerPort   - Basisport der zu verwendenden Schnittstelle */
/*           pPaket     - Adresse der Daten                           */
/*           iPaketSize - Gre der Daten                             */
/* Ausgabe : Fehlerstatus                                             */
/**********************************************************************/
INT trans_SendPaket( INT iSerPort, PBYTE pPaket, UINT iPaketSize )
{
  UINT uCRC, uAction, uTimeOut, uActByte;
  BYTE bData, bDataOk, bNakLo;     /* erstes Byte des NAK Returncodes */

  uTimeOut = 0xFFFF;
  uAction = SEND_RECEIVER;

  while( uTimeOut )
  {
    bDataOk = FALSE;                         /* Keine Daten vorhanden */
    if( ser_IsDataAvaiable( iSerPort ) )
      if( ser_ReadByte( iSerPort, &bData, 1, 0, 0 ) != SER_SUCCESS )
        return TRANS_TRANSERROR;
      else bDataOk = TRUE;                              /* oder doch? */

    if( bDataOk )
    {
      switch( uAction )
      {
        case SEND_RECEIVER:
             if( bData == ASCII_SYN )
             {
               uActByte = 0;
               uCRC = 0xFFFF;
               uAction = SEND_SENDING;
             }
        break;   /* Senden eines Paketes wird durch NAK unterbrochen! */
        case SEND_SENDING:
             if( bData == ASCII_NAK )  uAction = SEND_GETNAKLO;
        break;
        case SEND_GETNAKLO:
             bNakLo = bData;
             uAction = SEND_GETNAKHI;
        break;
        case SEND_GETNAKHI:
             return MAKEWORD( bData, bNakLo );

        case SEND_GETREPLY:
             if( bData == ASCII_ACK ) return TRANS_SUCCESS;
             if( bData == ASCII_NAK ) uAction = SEND_GETNAKLO;
             return TRANS_PROTOCOLL;
      }
    }
    else
    if( ser_IsWritingPossible( iSerPort ) )
    {
      switch( uAction )
      {
        case SEND_SENDING:
             if( ser_WriteByte( iSerPort,
                                pPaket[ uActByte ],
                                1,
                                0,
                                0 ) == SER_SUCCESS )
             {
                uCRC = CalcCRC( uCRC, pPaket[ uActByte ] );
                uActByte++;
                if( uActByte == iPaketSize )  uAction = SEND_PUTCRCLO;
                uTimeOut = 0xFFFF;
             }
             else return TRANS_TRANSERROR;
        break;
        case SEND_PUTCRCLO:
             if( ser_WriteByte( iSerPort,
                                LOBYTE( uCRC ),
                                1,
                                0,
                                0 ) == SER_SUCCESS )
                uAction = SEND_PUTCRCHI;
             else return TRANS_TRANSERROR;
        break;
        case SEND_PUTCRCHI:
             if( ser_WriteByte( iSerPort,
                                HIBYTE( uCRC ),
                                1,
                                0 ,
                                0 ) == SER_SUCCESS )
                 uAction = SEND_GETREPLY;
             else return TRANS_TRANSERROR;
        break;
        default: uTimeOut--; break;
      }
    }
  }
  return TRANS_TIMEOUT;
}

/**********************************************************************/
/* trans_ReceivePaket :Datenpaket von der Schnittstelle empfangen     */
/**------------------------------------------------------------------**/
/* Eingabe : iSerPort   - Basisport der zu verwendenden Schnittstelle */
/*           pPaket     - Adresse des Empfangsbuffers                 */
/*           iPaketSize - Gre des Buffers                           */
/* Ausgabe : Fehlerstatus                                             */
/**********************************************************************/
INT trans_ReceivePaket( INT iSerPort, PBYTE pPaket, UINT iPaketSize )
{
  UINT uCRC, uAction, uTimeOut, uActByte;
  BYTE bData, bDataOk, bCrcLo;           /* erstes Byte des CRC Codes */
  INT  iAckError;

  uTimeOut = 0xFFFF;
  uAction = RECEIVE_SIGNAL;

  uCRC = 0xFFFF;         /* wenigstens einmal auf Variablen zugreifen */
  uActByte = 0;

  while( uTimeOut )
  {
    bDataOk = FALSE;                         /* Keine Daten vorhanden */
    if( ser_IsDataAvaiable( iSerPort ) )
      if( ser_ReadByte( iSerPort, &bData, 1, 0, 0 ) != SER_SUCCESS )
        return TRANS_TRANSERROR;
      else bDataOk = TRUE;                              /* oder doch? */

    if( bDataOk )
    {
      switch( uAction )
      {
        case RECEIVE_RECEIVE:  
             uCRC = CalcCRC( uCRC, bData );
             pPaket[ uActByte++ ] = bData;
             if( uActByte == iPaketSize )  uAction = RECEIVE_GETCRCLO;
             uTimeOut = 0xFFFF;
        break;

        case RECEIVE_GETCRCLO:
             bCrcLo = bData;
             uAction = RECEIVE_GETCRCHI;
        break;
        case RECEIVE_GETCRCHI:
             if( uCRC == MAKEWORD( bData, bCrcLo ) )
                 uAction = RECEIVE_PUTACK;
             else
             {
               iAckError = TRANS_CRC;
               uAction = RECEIVE_PUTNAK;
             }
        break;
      }
    }
    else
    if( ser_IsWritingPossible( iSerPort ) )
    {
      switch( uAction )
      {
        case RECEIVE_SIGNAL:
             if( ser_WriteByte( iSerPort,
                                ASCII_SYN,
                                1,
                                0,
                                0 ) == SER_SUCCESS )
             {
                uCRC = 0xFFFF;
                uActByte = 0;
                uAction = RECEIVE_RECEIVE;
             }
             else return TRANS_TRANSERROR;
        break;
        case RECEIVE_PUTACK:
             if( ser_WriteByte( iSerPort,
                                ASCII_ACK,
                                1,
                                0,
                                0 ) == SER_SUCCESS )
                 return TRANS_SUCCESS;
             else return TRANS_TRANSERROR;

        case RECEIVE_PUTNAK:
             if( ser_WriteByte( iSerPort,
                                ASCII_NAK,
                                1,
                                0,
                                0 ) == SER_SUCCESS )
               uAction = RECEIVE_PUTNAKLO;
             else return TRANS_TRANSERROR;
        break;
        case RECEIVE_PUTNAKLO: 
             if( ser_WriteByte( iSerPort,
                                LOBYTE( iAckError ),
                                1,
                                0,
                                0 ) == SER_SUCCESS )
               uAction = RECEIVE_PUTNAKHI;
             else return TRANS_TRANSERROR;
        break;
        case RECEIVE_PUTNAKHI:
             if( ser_WriteByte( iSerPort,
                                HIBYTE( iAckError ),
                                1,
                                0,
                                0 ) == SER_SUCCESS )
               return iAckError;
             else return TRANS_TRANSERROR;
        default:  uTimeOut--;
      }
    }
  }
  return TRANS_TIMEOUT;
}

/**********************************************************************/
/* trans_ReceiveFile : Ganze Datei ber Schnittstelle empfangen       */
/**------------------------------------------------------------------**/
/* Eingabe : iSerPort - Basisport der zu verwendenden Schnittstelle   */
/* Ausgabe : Fehlerstatus                                             */
/**********************************************************************/
INT trans_ReceiveFile( INT iSerPort )
{ PAKET Paket;
  INT   iTimeOut;
  FILE *fHandle = NULL;
  UINT  uNr = 0;

  iTimeOut = 10;         /* Beim ersten Paket ein wenig lnger warten */
  while( iTimeOut )
  { INT err;
    err = trans_ReceivePaket( iSerPort,              /* Paket abholen */
                              ( PBYTE ) &Paket,
                              sizeof( Paket ) );
    if( err == TRANS_SUCCESS )
    {
      iTimeOut = 4;                              /* Timeout verkrzen */
      if( uNr != Paket.uNr )
      {
        if( fHandle ) fclose( fHandle );
        trans_SendNak( iSerPort, ( UINT )PAKET_SEQUENCE );
        return PAKET_SEQUENCE;
      }
      else
      {
        uNr++;
        printf("\r%d", uNr );
        switch( Paket.bType )
        {
          case PAKET_FILEOPEN:
               fHandle = fopen( ( PCHAR )Paket.bData, "wb" );
               if( !fHandle )
               {
                 trans_SendNak( iSerPort, _doserrno );
                 return _doserrno;
               }
          break;
          case PAKET_FILEDATA:
               if( fwrite( Paket.bData,
                           1,
                           Paket.uDataSize,
                           fHandle ) != Paket.uDataSize )
               {
                 trans_SendNak( iSerPort, _doserrno );
                 fclose( fHandle );
                 return _doserrno;
               }
          break;
          case PAKET_FILECLOSE:
               fclose( fHandle );
               return PAKET_SUCCESS;
          default:
               trans_SendNak( iSerPort, ( UINT )PAKET_UNKNOWN );
               return PAKET_UNKNOWN;
        }
      }
    }
    else  iTimeOut--;
  }
  return PAKET_TIMEOUT;
}

/**********************************************************************/
/* trans_SendFile : Ganze Datei ber Schnittstelle senden             */
/**------------------------------------------------------------------**/
/* Eingabe : iSerPort - Basisport der zu verwendenden Schnittstelle   */
/*           pName    - Name der zu versendenden Datei                */
/* Ausgabe : Fehlerstatus                                             */
/**********************************************************************/
INT trans_SendFile( INT iSerPort, PCHAR pName )
{
  FILE *fHandle;
  PAKET Paket;
  INT   iSize;

  fHandle = fopen( pName, "rb" );
  if( fHandle )
  { INT err;
    INT iCnt;
    INT iLen;
    UINT uNr = 0;

    iLen = strlen( pName );
    while( ( iLen > 0 ) &&
           ( pName[ iLen - 1 ] != '\\' ) &&
           ( pName[ iLen - 1 ] != ':' ) ) iLen--;
    paket_Build( &Paket,
                 PAKET_FILEOPEN,
                 ( PBYTE )&pName[ iLen ],
                 strlen( &pName[iLen] ) + 1,
                 uNr++ );

    iCnt = 25;
    printf("\r%d", uNr );
    do
    {
      iCnt--;
      err = trans_SendPaket( iSerPort,
                             ( PBYTE )&Paket,
                             sizeof( PAKET ) );
    } while( ( err == TRANS_TIMEOUT ) && ( iCnt ) );

    if( err != TRANS_SUCCESS )
    {
      fclose( fHandle );
      return err;
    }

    do
    {
      iSize = fread( Paket.bData, 1, PAKET_DATASIZE, fHandle );
      if( iSize )
      {
        paket_Build( &Paket,
                     PAKET_FILEDATA,
                     Paket.bData,
                     iSize,
                     uNr++ );
        iCnt = 10;
        printf("\r%d", uNr );
        do
        {
          iCnt--;
          err = trans_SendPaket( iSerPort,
                                 ( PBYTE )&Paket,
                                 sizeof( PAKET ) );
        } while ( ( err == TRANS_TIMEOUT ) && ( iCnt ) );
        if( err != TRANS_SUCCESS )
        {
          fclose( fHandle );
          return err;
        }
      }
    } while( iSize == PAKET_DATASIZE );

    fclose( fHandle );
    paket_Build( &Paket,
                 PAKET_FILECLOSE,
                 ( PBYTE )&pName[iLen],
                 strlen( &pName[iLen] ),
                 uNr++ );

    iCnt = 10;
    printf("\r%d", uNr );
    do
    {
      iCnt--;
      err = trans_SendPaket( iSerPort,
                             ( PBYTE )&Paket,
                             sizeof( PAKET ) );
    } while( ( err == TRANS_TIMEOUT ) && ( iCnt ) );
    return err;
  }
  return TRANS_NOFILE;
}

/**********************************************************************/
/* trans_PError : Fehlermeldung ausgeben                              */
/**------------------------------------------------------------------**/
/* Eingabe : iError - aufgetretener Fehler                            */
/**********************************************************************/
VOID trans_PError( INT iError )
{
  switch( iError )
  {
    case TRANS_SUCCESS:
         printf("bertragung OK!\n");                      break;
    case TRANS_TIMEOUT:
         printf("FEHLER: Byte-Timeout");                   break;
    case TRANS_TRANSERROR:
         printf("FEHLER: Genereller Fehler\n");            break;
    case TRANS_PROTOCOLL:
         printf("FEHLER: Ack oder NAck fehlt!\n");         break;
    case TRANS_CRC:
         printf("FEHLER: CRC-Fehler!");                    break;
    case TRANS_NOFILE:
         printf("FEHLER: Kann Datei nicht ffnen!\n");     break;
    case PAKET_SEQUENCE:
         printf("FEHLER: Ungltige Paket-Reihenfolge!\n"); break;
    case PAKET_UNKNOWN:
         printf("FEHLER: Unbekannte Paket-ID!\n");         break;
    case PAKET_TIMEOUT:
         printf("FEHLER: Paket-Timeout\n" );               break;
  }
}

/**********************************************************************/
/* H A U P T P R O G R A M M                                          */
/**********************************************************************/

VOID main(INT argc, PCHAR argv[] )
{
  INT iSerPort, iCom;
  LONG  lBaud;
  PCHAR pFileName;

  if( FindString( argv, "?", argc ) )
  {
    printf("Aufruf:\n" );
    printf("SERTRANS [SendeDatei] [-COM:port] [-BAUD:bdrate]\n");
    printf("comport  = 1 oder 2    (Default: 1 )\n");
    printf("baudrate = 50 - 115200 (Default: 9600)\n");
    exit(0);
  }

  if( GetArg( argc, argv, "-COM:", _int, &iCom, 1 ) )
  {
    if( iCom == 1 )
      iSerPort = SER_COM1;   /* Nur COM1 und COM2 sind "standardisiert" */
    else
    if( iCom == 2 )
      iSerPort = SER_COM2;   /* Nur COM1 und COM2 sind "standardisiert" */
    else
    {
      printf("Nicht untersttzter Com-Port!\n");
      exit(0);
    }
  }
  else
    iSerPort = SER_COM1;   /* Nur COM1 und COM2 sind "standardisiert" */

  if( !GetArg( argc, argv, "-BAUD:", _long, &lBaud, 1 ) )
      lBaud = 9600L;              /* Maximale Baudrate bei UART 8450A */
  if( lBaud > SER_MAXBAUD )
  {
    printf("Baudrate zu hoch!\n");
    printf("Maximum: %ld Bd\n", SER_MAXBAUD );
  }

  if( !ser_Init( iSerPort, lBaud,           /* Maximum bei UART 8450A */
           SER_LCR_8BITS | SER_LCR_1STOPBIT | SER_LCR_NOPARITY ) )
  {
    printf("Keine Schnittstelle!\n");
    exit( 0 );
  }
  
  ser_FIFOLevel( iSerPort, 0 );
  
  if( GetNArg( argc, argv, "-", &pFileName, 1 ) )
  {
    printf("Sending\n" );
    trans_PError( trans_SendFile( iSerPort, pFileName ) );
  }
  else
  {
    printf("Receiving\n");
    trans_PError( trans_ReceiveFile( iSerPort ) );
  }
}
